/*
 *      Copyright (c) 2018-2028, Chill Zhuang All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *  Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in the
 *  documentation and/or other materials provided with the distribution.
 *  Neither the name of the dreamlu.net developer nor the names of its
 *  contributors may be used to endorse or promote products derived from
 *  this software without specific prior written permission.
 *  Author: Chill 庄骞 (smallchill@163.com)
 */
package org.springblade.report.controller;

import cn.hutool.core.util.StrUtil;
import cn.hutool.http.ContentType;
import cn.hutool.http.HttpException;
import cn.hutool.http.HttpRequest;
import cn.hutool.http.HttpUtil;
import com.alibaba.fastjson2.JSON;
import com.alibaba.fastjson2.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.github.xiaoymin.knife4j.annotations.ApiOperationSupport;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import io.swagger.annotations.ApiParam;
import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springblade.config.properties.BiProperties;
import org.springblade.core.boot.ctrl.BladeController;
import org.springblade.core.log.annotation.ApiLog;
import org.springblade.core.mp.support.Condition;
import org.springblade.core.mp.support.Query;
import org.springblade.core.secure.BladeUser;
import org.springblade.core.tool.api.R;
import org.springblade.core.tool.api.ResultCode;
import org.springblade.core.tool.utils.Func;
import org.springblade.cube.service.DynamicTableService;
import org.springblade.cube.utils.CubeUtils;
import org.springblade.report.entity.DataApi;
import org.springblade.report.entity.ReportDatasource;
import org.springblade.report.service.IDataApiService;
import org.springblade.report.service.IReportDatasourceService;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import javax.servlet.http.HttpServletRequest;
import java.net.MalformedURLException;
import java.net.URL;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * 数据接口 控制器
 *
 * @author BladeX
 * @since 2021-04-17
 */
@Slf4j
@RestController
@AllArgsConstructor
@RequestMapping("report/dataapi")
@Api(value = "数据接口", tags = "数据接口接口 DataApiController")
public class DataApiController extends BladeController {

	private final IDataApiService dataApiService;
	private final IReportDatasourceService reportDatasourceService;
	private final DynamicTableService dynamicTableService;
	private final BiProperties biProperties;

	final static String ILLEGAL_CHARS = "!@#$%^&*()";

	/**
	 * 详情
	 */
	@GetMapping("/detail")
	@ApiOperationSupport(order = 1)
	@ApiOperation(value = "详情 detail", notes = "传入dataApi")
	public R<DataApi> detail(DataApi dataApi) {
		DataApi detail = dataApiService.getOne(Condition.getQueryWrapper(dataApi));
		if (detail == null) {
			return R.fail("数据不存在");
		}
		return R.data(detail);
	}

	/**
	 * 分页 数据接口
	 */
	@GetMapping("/list")
	@ApiOperationSupport(order = 2)
	@ApiOperation(value = "列表 list", notes = "传入dataApi")
	public R<List<DataApi>> list(DataApi dataApi, String apiTypes) {
		QueryWrapper<DataApi> queryWrapper = Condition.getQueryWrapper(dataApi);
		if (StrUtil.isNotBlank(apiTypes)) {
			String[] apiTypeArr = apiTypes.split(",");
			queryWrapper.lambda().in(DataApi::getApiType, Stream.of(apiTypeArr).map(Integer::valueOf).collect(Collectors.toList()));
		}
		if (StrUtil.isNotBlank(dataApi.getApiName())) {
			queryWrapper.lambda().like(DataApi::getApiName, dataApi.getApiName());
		}
		List<DataApi> list = dataApiService.list(queryWrapper);
		return R.data(list);
	}

	/**
	 * 自定义分页 数据接口
	 */
	@GetMapping("/page")
	@ApiOperationSupport(order = 3)
	@ApiOperation(value = "分页 page", notes = "传入dataApi")
	public R<IPage<DataApi>> page(DataApi dataApi, Query query) {
		QueryWrapper<DataApi> queryWrapper = Condition.getQueryWrapper(dataApi);
		return R.data(dataApiService.page(Condition.getPage(query), queryWrapper.lambda().orderByDesc(DataApi::getCreateTime)));
	}

	/**
	 * 新增或修改 数据接口
	 */
	@PostMapping("/submit")
	@ApiOperationSupport(order = 6)
	@ApiOperation(value = "新增或修改", notes = "传入dataApi")
	@ApiLog("API数据接口")
	public R<DataApi> submit(@RequestBody String value) throws Exception {
		String decoderStr = new String(Base64.getDecoder().decode(value), "utf-8");
		DataApi dataApi = JSON.parseObject(decoderStr, DataApi.class);
		if (StrUtil.containsAny(dataApi.getApiName(), ILLEGAL_CHARS.split(""))){
			return R.fail("名字包含特殊字符，请修改");
		}
		if (StrUtil.containsAny(dataApi.getApiKey(), ILLEGAL_CHARS.split(""))){
			return R.fail("apikey包含特殊字符，请修改");
		}
		if (dataApi.getApiName().length() > 50 || dataApi.getApiKey().length() > 50) {
			return R.fail("名字和apiKey长度不能大于50");
		}
		if (null == dataApi.getApiType()) {
			return R.fail("API类型不能为空");
		}
		if (StrUtil.isBlank(dataApi.getApiKey())) {
			return R.fail("api关键词不可以为空");
		}
		if (2 == dataApi.getApiType() && !isURL(dataApi.getApiUrl())) {
			return R.fail("api地址配置不正确");
		}
		if (2 != dataApi.getApiType() && null == dataApi.getDatasourceId()) {
			return R.fail("api数据源不可以为空");
		}
		if (1 == dataApi.getApiType() && StrUtil.isBlank(dataApi.getQuerySql())) {
			return R.fail("常规接口查询sql不可以为空");
		}
		if (1 == dataApi.getApiType() && checkSqlKeywords(dataApi.getQuerySql())) {
			return R.fail("查询sql包含非法关键字");
		}
		if (3 == dataApi.getApiType() && StrUtil.isBlank(dataApi.getQueryParam())) {
			return R.fail("业务参数不能为空");
		}
		//新增：设置创建人ID
		if(dataApi.getId() == null) {
			dataApi.setCreateUserId(getUser().getUserId());
		}
		//判断apiKey是否存在
		LambdaQueryWrapper<DataApi> countQueryWrapper = Wrappers.<DataApi>lambdaQuery().eq(DataApi::getApiKey, dataApi.getApiKey()).ne(dataApi.getId() != null, DataApi::getId, dataApi.getId());
		long count = dataApiService.count(countQueryWrapper);
		if(count > 0) {
			return R.fail("关键字不能重复");
		}
		if (1 == dataApi.getApiType()) {
			dataApi.setApiUrl("/report/dataapi/open/" + dataApi.getApiKey());
		}else if(3 == dataApi.getApiType()) {
			dataApi.setApiUrl("/data-api/open/" + dataApi.getApiKey());
		}
		dataApi.setCreateTime(LocalDateTime.now());
		dataApiService.saveOrUpdate(dataApi);
		return R.data(dataApi);
	}

	/**
	 * 删除 数据接口
	 */
	@PostMapping("/remove")
	@ApiOperationSupport(order = 7)
	@ApiOperation(value = "删除 remove", notes = "传入ids")
	@ApiLog("API数据接口")
	public R remove(@ApiParam(value = "主键集合", required = true) @RequestParam String ids) {
		return R.status(dataApiService.removeByIds(Func.toLongList(ids)));
	}

	/**
	 * 对外暴露的接口
	 * @param apiKey
	 * @param apiParam
	 * @param request
	 * @return
	 */
	@GetMapping("/open/{apiKey}")
	@ApiOperation(value = "执行常规类接口", notes = "传入apiKey【说明：第三方接口传递apiParam(接口参数)】")
	public R executeDataApi(@PathVariable("apiKey") String apiKey, String apiParam, HttpServletRequest request) {
		log.info("executeDataApi", apiKey + request.getParameterMap());
		if (StrUtil.isBlank(apiKey)) {
			return R.fail("接口关键词不可以为空");
		}
		DataApi dataApi = dataApiService.getOneByApiKey(apiKey);
		if (null == dataApi) {
			return R.fail("接口不存在，请检查接口关键词");
		}
		if(dataApi.getApiType() == 2) { //第三方接口
			//获取header
			Map<String, String> headers = null;
			if(!StrUtil.isEmpty(dataApi.getApiHeaders())) {
				headers = JSON.parseObject(dataApi.getApiHeaders(), Map.class);
			}
			JSONObject queries = StrUtil.isEmpty(apiParam) ? null : JSON.parseObject(apiParam);
			String postJson = StrUtil.isEmpty(apiParam) ? "{}" : apiParam;
			try {
				HttpRequest httpRequest = null;
				if("post".equalsIgnoreCase(dataApi.getApiMethod())) {
					httpRequest = HttpUtil.createPost(dataApi.getApiUrl());
					if(dataApi.getApiParamType() == 0) {//JSON方式传参数
						httpRequest.body(postJson, ContentType.JSON.getValue());
					}else {
						httpRequest.form(queries);
					}
				}else {
					httpRequest = HttpUtil.createGet(dataApi.getApiUrl()).form(queries);
				}
				//请求数据
				String retBody = httpRequest.setConnectionTimeout(biProperties.getDataApiTimeoutSecs() * 1000).setReadTimeout(biProperties.getDataApiTimeoutSecs() * 1000).addHeaders(headers).execute().body();
				return R.data(retBody);
			} catch (HttpException e) {
				log.error(e.getMessage(), e);
				return R.fail("API接口请求异常，请检查地址");
			}
		}else {
			if (null == dataApi.getDatasourceId()) {
				return R.fail("接口未配置数据源");
			}
			if (1 != dataApi.getApiType()) {
				return R.fail("接口类型配置错误");
			}
			if (null != dataApi.getLoginRequired() && 1 == dataApi.getLoginRequired()) {
				BladeUser user = getUser();
				if (null == user) {
					return R.fail("未登录");
				}
			}
			String querySql = dataApi.getQuerySql();
			Map<String, String[]> maps = request.getParameterMap();
			String regex = "\\@(.*?)\\@";
			Pattern p = Pattern.compile(regex);
			Matcher m = p.matcher(querySql);
			//Map<String, Object> params = new HashMap<>();
			List<String> paramList = new ArrayList<>();
			while (m.find()) {
				String key = m.group(1);
				if (StrUtil.isNotBlank(key)) {
					paramList.add(key);
				}
			}
			String regex2 = "\\@@(.*?)\\@@";
			Pattern p2 = Pattern.compile(regex2);
			Matcher m2 = p2.matcher(querySql);
			while (m2.find()) {
				String key = m2.group(1);
				if (StrUtil.isNotBlank(key)) {
					paramList.add(key);
				}
			}
			ReportDatasource datasource = reportDatasourceService.getById(dataApi.getDatasourceId());
			if (null == datasource) {
				return R.fail("接口数据源配置错误");
			}
			for (String key : paramList) {
				String value = "";
				if (maps.size() > 0 && null != maps.get(key)) {
					value = maps.get(key)[0];
					if (checkSqlKeywords(value)) {
						return R.fail("入参包含非法字符");
					}
				}
				if (querySql.contains("@@" + key + "@@")) {
					if (StrUtil.isNotBlank(value)) {
						querySql = querySql.replaceAll("@@" + key + "@@", value);
					} else {
						querySql = querySql.replaceAll("@@" + key + "@@", "1=1 ");
					}
				} else if (querySql.contains("@" + key + "@")) {
					querySql = querySql.replace("@" + key + "@", "'" + value + "'");
				}
			}

			try {
				List<Map<String, Object>> list = dynamicTableService.selectList(datasource, querySql);
				return R.data(list);
			} catch (Exception e) {
				log.error(e.getMessage(), e);
				return R.data(ResultCode.FAILURE.getCode(), e.getMessage(), "请检查SQL语法是否有误");
			}
		}

	}
	
	@PostMapping("/check")
	@ApiOperation(value = "检测ApiKey是否存在接口", notes = "传入apiKey")
	public R<Boolean> checkApiKey(@RequestParam String apiKey){
		QueryWrapper queryWrapper = new QueryWrapper();
		queryWrapper.eq("api_key", apiKey);
		DataApi dataApi = dataApiService.getOne(queryWrapper);
		if (null == dataApi) {
			return R.data(false, "接口不存在");
		}else {
			return R.data(true, "接口存在");
		}
	}

	boolean checkSqlKeywords(String sql) {
		return CubeUtils.checkSqlInject(sql);
	}

	boolean isURL(String str) {
		try {
			new URL(str);
			return true; // 如果没有抛出异常，则说明是一个URL
		} catch (MalformedURLException e) {
			return false; // 如果抛出异常，则说明不是一个URL
		}
	}

}
